/*
MatlabWrapper

Andrej Karpathy
August 27, 2012
BSD Licence
*/
#include "stdafx.h"
#include "MatlabWrapper.h"
#include <iostream>
#include <map>
 
using namespace std; 
MatlabWrapper::MatlabWrapper() {
  initted_ = false; 
  verbose_ = true;
} 

MatlabWrapper::~MatlabWrapper() {
  engClose(ep_); // close down the engine
}

bool MatlabWrapper::init() {
  
	if (!(ep_ = engOpen("-nosplash "))) {
		fprintf(stderr, "\nCan't start MATLAB engine.\n");
		return false;
	}else{
		printf("Start MATLAB engine.\n");
	}
	
	// connect Matlab engine to a char buffer for output
	buffer_[MATLAB_BUF_SIZE - 1] = '\0';
	engOutputBuffer(ep_, buffer_, MATLAB_BUF_SIZE - 1);
	//engOutputBuffer(ep_, NULL, 0);
	initted_ = true;
	return true;
}

void MatlabWrapper::sendMat(const string name, const Eigen::MatrixXf &mat) {
  
  int N = mat.rows();
  int D = mat.cols();
  if(N <= 0 || D <= 0) {
    fprintf(stderr, "\nWow, weird matrix is of size %dx%d. Ignoring sendMat request.\n", N, D);
    return;
  }
  if(!initted_) { init(); }
  
  mxArray *X = mxCreateNumericMatrix(N, D, mxSINGLE_CLASS, mxREAL);
  memcpy((float *)mxGetData(X), mat.data(), sizeof(float)*N*D);
	engPutVariable(ep_, name.c_str(), X);
	mxDestroyArray(X); // the engine takes ownership of a copy
}

void MatlabWrapper::sendMat(const string name, const Eigen::MatrixXd &mat) 
{

	int N = mat.rows();
	int D = mat.cols();
	if(N <= 0 || D <= 0) {
		fprintf(stderr, "\nWow, weird matrix is of size %dx%d. Ignoring sendMat request.\n", N, D);
		return;
	}
	if(!initted_) { init(); }

	mxArray *X = mxCreateNumericMatrix(N, D, mxDOUBLE_CLASS, mxREAL);
	memcpy((float *)mxGetData(X), mat.data(), sizeof(double)*N*D);
	engPutVariable(ep_, name.c_str(), X);
	mxDestroyArray(X); // the engine takes ownership of a copy
}

bool MatlabWrapper::receiveMat(const string name, Eigen::MatrixXi &mat) const
{

	mxArray *result = NULL;
	if ((result = engGetVariable(ep_, name.c_str())) == NULL) {
		printf("Variable %s is not defined! Ignoring request.\n", name.c_str());
		throw "Can only receive 2D matrix back! Exitting.";
		return false;
	}

	int number_of_dims; 
	number_of_dims = mxGetNumberOfDimensions(result);
	if(number_of_dims != 2) {
		printf("Can only receive 2D matrix back! Exitting.\n");
		mxDestroyArray(result);

		throw "Can only receive 2D matrix back! Exitting.";
		return false;
	}

	const mwSize *dim_array = mxGetDimensions(result);
	int N= dim_array[0];
	int D= dim_array[1];

	mxClassID classType= mxGetClassID(result);

	

	if( classType == mxUINT8_CLASS ) { 
		
		mat.resize(N, D);
		unsigned char *tmpData=new unsigned char[N*D];
		memcpy(tmpData, (unsigned char *)mxGetData(result), sizeof(unsigned char)*N*D);
		
		int * innerData = mat.data();
		for(int i=0 ; i < N*D ; i++){
			innerData[i] = tmpData[i];
		}

		delete [] tmpData;
	} else if( classType == mxINT32_CLASS ) { 

		mat.resize(N, D);
		int *tmpData=new int[N*D];
		memcpy(tmpData, (int *)mxGetData(result), sizeof(int)*N*D);

		int * innerData = mat.data();
		for(int i=0 ; i < N*D ; i++){
			innerData[i] = tmpData[i];
		}

		delete [] tmpData;
	} else if(classType == mxDOUBLE_CLASS) {

		Eigen::MatrixXd mat_2;

		mat_2.resize(N, D);
		memcpy(mat_2.data(), (double *)mxGetData(result), sizeof(double)*N*D);

		mat = mat_2.cast<int>();

		printf("**auto conv double to int %s\n", name.c_str());
	} else {
		printf("Error! Variable %s is not uin8/int32. Ignoring.\n", name.c_str());
	}

	mxDestroyArray(result); // we own this newly allocated mxArray. Destroy it

	return true;
}
bool MatlabWrapper::receiveMat(const string name, Eigen::MatrixXd &mat) const
{

	mxArray *result = NULL;
	if ((result = engGetVariable(ep_, name.c_str())) == NULL) {
		printf("Variable %s is not defined! Ignoring request.\n", name.c_str());
		throw "Can only receive 2D matrix back! Exitting.";
		return false;
	}

	int number_of_dims; 
	number_of_dims = mxGetNumberOfDimensions(result);
	if(number_of_dims != 2) {
		printf("Can only receive 2D matrix back! Exitting.\n");
		mxDestroyArray(result);
		throw "Can only receive 2D matrix back! Exitting.";
		return false;
	}

	const mwSize *dim_array = mxGetDimensions(result);
	int N= dim_array[0];
	int D= dim_array[1];

	mxClassID classType= mxGetClassID(result);
	if(classType == mxSINGLE_CLASS) { 
		mat.resize(N, D);
		memcpy(mat.data(), (float *)mxGetData(result), sizeof(float)*N*D);

	} else if(classType == mxDOUBLE_CLASS) {
		mat.resize(N, D);
		memcpy(mat.data(), (double *)mxGetData(result), sizeof(double)*N*D);
	} else {
		printf("Error! Variable %s is not single/double. Ignoring.\n", name.c_str());
	}

	mxDestroyArray(result); // we own this newly allocated mxArray. Destroy it

	return true;
}

bool MatlabWrapper::receiveMat(const string name, Eigen::MatrixXf &mat) const {
  
  mxArray *result = NULL;
  if ((result = engGetVariable(ep_, name.c_str())) == NULL) {
    printf("Variable %s is not defined! Ignoring request.\n", name.c_str());
	throw "Can only receive 2D matrix back! Exitting.";
    return false;
  }
  
  int number_of_dims; 
  number_of_dims = mxGetNumberOfDimensions(result);
  if(number_of_dims != 2) {
    printf("Can only receive 2D matrix back! Exiting.\n");
    mxDestroyArray(result);
	throw "Can only receive 2D matrix back! Exitting.";
    return false;
  }
  
  const mwSize *dim_array = mxGetDimensions(result);
  int N= dim_array[0];
  int D= dim_array[1];
  
  mxClassID classType= mxGetClassID(result);
  if(classType == mxSINGLE_CLASS) { 
    mat.resize(N, D);
    memcpy(mat.data(), (float *)mxGetData(result), sizeof(float)*N*D);
    
  }else if(classType == mxDOUBLE_CLASS ){
   
	if(verbose_)
		printf("cvt %s double to single!\n", name.c_str());
    
	Eigen::MatrixXd proxy(N, D);
    memcpy(proxy.data(), (double *)mxGetData(result), sizeof(double)*N*D);
    mat = proxy.cast<float>();
    
  }else{

    printf("Error! Variable %s is not single/double. Ignoring.\n", name.c_str());
  }
  
  mxDestroyArray(result); // we own this newly allocated mxArray. Destroy it
  return true;
}

void MatlabWrapper::exec(const string cmd, const bool waitForKey) const {

  engEvalString(ep_, cmd.c_str());
  printf("-------------\n[MATLAB]:exec[%s]\n%s\n-------------\n",cmd.c_str(), buffer_);

  if(waitForKey) {
    printf("Hit return to continue\n"); 
    fgetc(stdin);
  }
}

void MatlabWrapper::interact() {
  
  printf("Started interactive MATLAB session! Enter 'qq' to exit.\n");
  char str[MATLAB_BUF_SIZE + 1];
  while(true) {
    
    printf(">> ");
    fgets(str, MATLAB_BUF_SIZE, stdin);
    if(strcmp(str, "qq\n") == 0) break;
    
    engEvalString(ep_, str);
    printf("%s\n", buffer_);
  }
}



